// APropList.cpp 
//
	/* *****************************
	 *
	 * Read the Apple Plist files from the iPhone and converts them to/from binary format
	 * Requires CoreFoundation.dll and its support dlls from a iTunes install.
	 * These are found in 
	 * C:\Program Files\Common Files\Apple\Mobile Device Support\bin\
	 *
	 * *****************************/

#include "APropList.h"


using namespace std;

int main(int argc,char * argv[])
{
ACF_Interface* ACFi;

//
//  Function Return Values
//  
CFIndex ofile_BytesWritten;
CFStringRef ifile_CFStringRef, ofile_CFStringRef;
CFURLRef ifile_CFURLRef, ofile_CFURLRef;
CFReadStreamRef ifile_CFReadStreamRef;
CFWriteStreamRef ofile_CFWriteStreamRef;
CFPropertyListFormat ifile_PlistFormat, ofile_PlistFormat;
CFStringRef errorString ;
CFPropertyListRef PlistRef;

if(argc != 3) 
{
	usage();
	return EXIT_FAILURE;
}
try {
ACFi = new ACF_Interface();
}
catch (int ) {
	return EXIT_FAILURE;
}
//
//  Create a CFStringRef for the input & iutput files
//

	ifile_CFStringRef = ACFi->CFStringCreateWithCString(NULL, (const char *) argv[1],kCFStringEncodingWindowsLatin1);

	if(!ifile_CFStringRef)
	{
		cout <<  "Build of ifile_CFStringRef failed" << endl;
		return EXIT_FAILURE;
	}

	ofile_CFStringRef = ACFi->CFStringCreateWithCString(NULL,(const char *) argv[2],kCFStringEncodingWindowsLatin1);

	if(!ofile_CFStringRef)
	{
		cout <<  "Build of ofile_CFStringRef failed" << endl;
		return EXIT_FAILURE;
	}
//
//  Create CFURLRef(s) from the CFStringRef(s)
//
	ifile_CFURLRef = ACFi->CFURLCreateWithFileSystemPath(NULL, ifile_CFStringRef, kCFURLWindowsPathStyle,FALSE);
	if(!ifile_CFURLRef)
	{
		cout <<  "Build of ifile_CFURLRef failed" << endl;
		return EXIT_FAILURE;
	}
	ofile_CFURLRef = ACFi->CFURLCreateWithFileSystemPath(NULL, ofile_CFStringRef, kCFURLWindowsPathStyle,FALSE);
	if(!ofile_CFURLRef)
	{
		cout <<  "Build of ofile_CFURLRef failed" << endl;
		return EXIT_FAILURE;
	}
//
//  Create a CFReadStreamRef for the input file
//
	ifile_CFReadStreamRef = ACFi->CFReadStreamCreateWithFile(NULL, ifile_CFURLRef);
	if(!ifile_CFReadStreamRef)
	{
		cout <<  "Build of ifile_CFReadStreamRef failed" << endl;
		return EXIT_FAILURE;
	}
//
// Create CFWriteStreamRef for the output file
//
	ofile_CFWriteStreamRef = ACFi->CFWriteStreamCreateWithFile(NULL, ofile_CFURLRef);
	if(!ofile_CFWriteStreamRef)
	{
		cout <<  "Build of ofile_CFWriteStreamRef failed" << endl;
		return EXIT_FAILURE;
	}
//
//  open the ReadStream
//
	if(!(ACFi->CFReadStreamOpen(ifile_CFReadStreamRef)))
	{
		cout <<  "Unable to open input file (CFReadStreamOpen failed)" << endl;
		return EXIT_FAILURE;
	}
//
//  open the WriteStream
//
	if(!(ACFi->CFWriteStreamOpen(ofile_CFWriteStreamRef)))
	{
		cout <<  "Unable to open output file (CFWriteStreamOpen failed)" << endl;
		return EXIT_FAILURE;
	}
//
//  Create the Plist Object
//
	PlistRef  = ACFi->CFPropertyListCreateFromStream(NULL,ifile_CFReadStreamRef,0,
													kCFPropertyListMutableContainersAndLeaves,&ifile_PlistFormat,&errorString);
   if(!(PlistRef))
	{
		cout <<  "Unable Create PropertyList Object -- probably invalid plist file (CFPropertyListCreateFromStream failed)" << endl;
		return EXIT_FAILURE;
	}
//
//  Close the ReadSteam
//
   ACFi->CFReadStreamClose(ifile_CFReadStreamRef);
   
//
//  Verify the property list
//
   	if(!ACFi->CFPropertyListIsValid(PlistRef,ifile_PlistFormat))
	{
		cout <<  "PropertyList is invalid (CFPropertyListIsValid failed)" << endl;
		return EXIT_FAILURE;
	}
//
//   Write out the PropertyList as an XML file
//
	if(ifile_PlistFormat == kCFPropertyListXMLFormat_v1_0)
		ofile_PlistFormat = kCFPropertyListBinaryFormat_v1_0;
	else
		ofile_PlistFormat = kCFPropertyListXMLFormat_v1_0;

  	ofile_BytesWritten  = ACFi->CFPropertyListWriteToStream(PlistRef,ofile_CFWriteStreamRef, ofile_PlistFormat, &errorString);
	if(!ofile_BytesWritten)
	{
		cout <<  "Write of PropertyList failed -- don't know why if you got here ?? (CFPropertyWriteToStream failed)" << endl;
		return EXIT_FAILURE;
	}
//
//  Close the WriteSteam
//
      ACFi->CFWriteStreamClose(ofile_CFWriteStreamRef);
   
//
//  quit
//
	return EXIT_SUCCESS;
}
void usage()
{
	cout <<  "AProplist.exe" << endl;
	cout <<  "Convert an Apple binary formated plist file to XML" << endl;
	cout <<  "--------" << endl;
	cout <<  "usage: ApplePlist input.file.name outputfile.name" << endl;
	return ;
}
