#include "RecoveryInterface.h"

static am_recovery_device *g_recoveryDevice;

void recovery_progress_callback(unsigned int progress_number, unsigned int opcode)
{
	fprintf(stderr, "Recovery progress callback...\n");
	if (opcode == 9) {
		sendCommandToDevice(g_recoveryDevice, CFSTR("setenv boot-args rd=md0 -v"));
	}
}

int recovery_restore(string *args, struct shell_state *sh)
{
	
	mach_error_t ret;
	
	if (args[1] == "")
	{
		ifNotQuiet cout << "args[1] must be RestoreBundlePath" << endl;
		return SHELL_CONTINUE;
	}
	
	D("Building CFMutableDictionary");
	CFMutableDictionaryRef opts;
	D("Getting AMRestoreCreateDefaultOptions");
	opts = AMRestoreCreateDefaultOptions(kCFAllocatorDefault);
	CFStringRef value = CFStringCreateWithCString(kCFAllocatorDefault, args[1].c_str(),
												  kCFStringEncodingMacRoman);
	CFDictionarySetValue(opts, CFSTR("RestoreBundlePath"), value );
        
	//    describe255(opts);
        
	g_recoveryDevice = sh->recovery_dev;	
	ret = AMRestorePerformRecoveryModeRestore(	sh->recovery_dev, opts,
							(void *)recovery_progress_callback, NULL );
	CFRelease(value);
	ifVerbose cout 	<< "AMRestorePerformRecoveryModeRestore: " << ret << endl;
	
	return SHELL_CONTINUE;
}

int recovery_grestore(string *args, struct shell_state *sh)
{
	int ret;
	char* Ramdisk = NULL;
	char* KernelCache = NULL;
	string BundlePath;
	struct stat statbuffer;
	PHONEFIRMWARE pfw[] = {
//	  Ver           Ramdisk              KernelCache
 	{0x00000111,"022-3629-9.dmg","kernelcache.release.s5l8900xrb"},
 	{0x00000111,"022-3630-9.dmg","kernelcache.release.s5l8900xrb"},
 	{0x00000102,"009-7698-4.dmg","kernelcache.release.s5l8900xrb"},
 	{0x00000101,"009-7662-6.dmg","kernelcache.release.s5l8900xrb"},
 	{0x00000100,"694-5259-38.dmg","kernelcache.restore.release.s5l8900xrb"},
 	{0x00000000,"NoTfOuND","NoTfOuND"}
	};
	
	if (args[1] == "")
	{
		ifNotQuiet cout << "args[1] must be RestoreBundlePath" << endl;
		return SHELL_CONTINUE;
	}
	// check for trailing /
	if (args[1].length() != 0)
	{
	if ( args[1].at(args[1].length() - 1) != '/' )
		args[1] = args[1] + '/';
	}
#ifdef WIN32
	makeWIN32Path(args[1],&BundlePath,sh); 
#else
	BundlePath = args[1];
#endif	
	D("Looking for a usable Ramdisk/KernelCache pair");

		for (int i = 0; pfw[i].FWVersion; i++) {
			string Ramdiskstring = BundlePath + pfw[i].FWRamdisk;
			string KernelCachestring = BundlePath + pfw[i].FWKernelCache;
			if ((!(stat(Ramdiskstring.c_str(),&statbuffer))) && (!(stat(KernelCachestring.c_str(),&statbuffer)))) {
				Ramdisk = strdup(pfw[i].FWRamdisk);
				KernelCache = strdup(pfw[i].FWKernelCache);
			}
		}
		
	if(Ramdisk == NULL) {
		cout << "Unable to locate a valid firmware bundle for grestore" << endl;
		return SHELL_CONTINUE;
		}

	ifVerbose cout << "Found Firmware bundle" << Ramdisk << "/" << KernelCache << endl;
#ifdef WIN32
	SetCurrentDirectory(BundlePath.c_str());
	string temp = Ramdisk;
#else
	string temp = args[1] + Ramdisk;
#endif
	ifNotQuiet cout << "Bundle Name "  << temp << endl;
	ifNotQuiet cout << "Entering G-Restore ... " << endl;
	// Send ramdisk to the phone
	ifVerbose cout << "Sending ramdisk '"<< temp << "'" << endl;
	ret = sendFileToDevice(sh->recovery_dev, CFStringCreateWithCString(NULL, temp.c_str(), kCFStringEncodingASCII));
	ifVerbose cout << "sendFileToDevice: " << ret << endl;
	if (ret != 0) {
		ifNotQuiet cout << "Send ramdisk failed.  Aborting." << endl;
		free(Ramdisk);
		free(KernelCache);
		return SHELL_CONTINUE;
	}
	
	// load ramdisk on the phone
	ifVerbose cout << "Loading ramdisk." << endl;
	ret = sendCommandToDevice(sh->recovery_dev, CFSTR("ramdisk"));
	ifVerbose cout << "sendCommandToDevice 'ramdisk': " << ret << endl;
	if (ret != 0) {
		ifNotQuiet cout << "Load ramdisk failed.  Aborting." << endl;
		free(Ramdisk);
		free(KernelCache);
		return SHELL_CONTINUE;
	}
	

#ifdef WIN32
	temp = KernelCache;
#else
	temp = args[1]+ KernelCache;
#endif
	// Send the kernelcache
	ifVerbose cout << "Sending kernelcache '"<< temp << "'" << endl;
	ret = sendFileToDevice(sh->recovery_dev, CFStringCreateWithCString(NULL, temp.c_str(), kCFStringEncodingASCII));
	ifVerbose cout << "sendFileToDevice: " << ret << endl;
	if (ret != 0) {
		ifNotQuiet cout << "Send kernelcache failed.  Aborting." << endl;
		free(Ramdisk);
		free(KernelCache);
		return SHELL_CONTINUE;
	}
#ifdef WIN32
	makeWIN32Path(sh->local_path,&BundlePath,sh);
	SetCurrentDirectory(BundlePath.c_str());
#endif
	//check for custom bootargs
	if( args[2] != "" )
		D("grestore does not yet except custom bootargs.");
	
	ifVerbose cout << "Setting bootargs: setenv boot-args rd=md0 -v" << endl;
	ret = sendCommandToDevice(sh->recovery_dev, CFSTR("setenv boot-args rd=md0 -v"));
	ifVerbose cout << "sendCommandToDevice: " << ret << endl;
	if (ret != 0) {
		ifNotQuiet cout << "Setevn command failed.  Aborting." << endl;
		free(Ramdisk);
		free(KernelCache);
		return SHELL_CONTINUE;
	}
	
	ifVerbose cout << "Loading kernelcache (command: bootx)" << endl;
	ret = sendCommandToDevice(sh->recovery_dev, CFSTR("bootx"));
	ifVerbose cout << "sendCommandToDevice: " << ret << endl;	
	if (ret != 0) {
		ifNotQuiet cout << "failed (" << ret << ") Aborting." << endl;
		free(Ramdisk);
		free(KernelCache);
		return SHELL_CONTINUE;
	}
	
	ifNotQuiet cout << "G-Restore Completed.  \nExit iPHUC or reconnect in order to continue in restore mode." << endl;
	free(Ramdisk);
	free(KernelCache);
	
	return SHELL_CONTINUE;
}

int recovery_filecopytophone(string *args, struct shell_state *sh)
{
	mach_error_t retval = sendFileToDevice(	sh->recovery_dev,
						CFStringCreateWithCString(NULL, args[1].c_str(), kCFStringEncodingASCII) );
	
	ifNotQuiet cout	<< "filecopytophone: " << retval << endl;
	//cout << "filecopytophone: " << filecopytophone(sh->recovery_dev, cline.c_str(), 0x09000000) << " bytes copied" << endl;
	
	return SHELL_CONTINUE;
}

int recovery_serial(string *args, struct shell_state *sh)
{
	ifNotQuiet cout << "Restarting in serial mode." << endl;
        
	D("bgcolor 255 255 0");
	sendCommandToDevice(sh->recovery_dev, CFSTR("bgcolor 255 255 0"));
	D("setenv debug-uarts 1");
	sendCommandToDevice(sh->recovery_dev, CFSTR("setenv debug-uarts 1"));
	D("saveenv");
	sendCommandToDevice(sh->recovery_dev, CFSTR("saveenv"));
	D("reboot");
	sendCommandToDevice(sh->recovery_dev, CFSTR("reboot"));
	
	//im ASSUMING we dont want to kill the app here, but that we want to
	// at least exit recovery mode (?)
	
	ifNotQuiet cout << "Please restart iPHUC or reconnect in order to issue serial commands." << endl;
	
	return SHELL_CONTINUE;
}

int recovery_cmd(string *args, struct shell_state *sh)
{
	
	if ( args[1] == "" )
	{
		ifNotQuiet cout << "Please provide a command to send to the phone." << endl;
		return SHELL_CONTINUE;
	}
	
	mach_error_t retval = sendCommandToDevice(	sh->recovery_dev,
							CFStringCreateWithCString(NULL, args[1].c_str(), kCFStringEncodingASCII) );
	
	ifVerbose cout	<< "cmd: " << args[1] << ": " << retval << endl; 
	
	return SHELL_CONTINUE;
}

int recovery_exit(string *args, struct shell_state *sh)
{
	D("exiting iphuc");
	exit(0);
}

int recovery_disconnect(string *args, struct shell_state *sh)
{
	return SHELL_WAIT;
}
#ifdef WIN32
int makeWIN32Path(string path, string *WIN32Path, struct shell_state *sh)
{

	int retc;
	WIN32Path->clear();
	string temp = path;
	if(temp.length() > 10) {					// "file:///[a-zA-Z]:/" is shortest absolute path
		if(temp.substr(0,8) != "file:///") {	// not an absolute path
			if(temp.substr(0,1) != "/")			// add a "/" if its not there
				temp.insert(0,"/");
			temp.insert(0,sh->local_path);		// add in current Directory for relative Path 
		}
	} else {
	if(temp.substr(0,1) != "/")			// add a "/" if its not there
		temp.insert(0,"/");
	temp.insert(0,sh->local_path);		// add in current Directory for relative Path 
	}
	char Path[MAX_PATH] = " ";
	DWORD sPath = MAX_PATH;
	retc = PathCreateFromUrl(temp.c_str(),Path,&sPath, DWORDNULL);
	WIN32Path->append(Path);
	return retc;
}	
#endif
